@@ -4,12 +4,15 @@ import os
 import pprint
 import re
 import urllib.parse
+import warnings
 from collections import defaultdict
 from concurrent.futures import ThreadPoolExecutor
 from copy import deepcopy
+from getpass import getpass
 from io import BytesIO
 from pathlib import Path
 from shutil import copyfileobj
+from threading import BoundedSemaphore
 from typing import Optional, List, Iterable, Union, Tuple, Any, TYPE_CHECKING
 from urllib.parse import urlparse
 
@@ -17,6 +20,7 @@ import numpy as np
 import pandas as pd
 import requests
 from PIL import Image
+from tqdm import tqdm
 
 from metaspace.image_processing import clip_hotspots
 from metaspace.types import (
@@ -190,16 +194,28 @@ def multipart_upload(local_path, companion_url, file_type, headers={}):
         resp_data = send_request(url, max_retries=2)
         return resp_data['url']
 
-    def upload_part(presigned_url, data):
-        resp_data = send_request(
-            presigned_url,
-            'PUT',
-            data=data,
-            headers=headers,
-            return_headers=True,
-            max_retries=2,
-        )
-        return resp_data['ETag']
+    def upload_part(part, data):
+        try:
+            max_retries = 3
+            for i in range(max_retries):
+                try:
+                    presigned_url = sign_part_upload(key, upload_id, part)
+                    resp_data = send_request(
+                        presigned_url,
+                        'PUT',
+                        data=data,
+                        headers=headers,
+                        return_headers=True,
+                        max_retries=2,
+                    )
+                    return resp_data['ETag']
+                except Exception as ex:
+                    if i == max_retries - 1:
+                        raise
+                    else:
+                        print(f'Part {part} failed with error: {ex}. Retrying...')
+        finally:
+            semaphore.release()
 
     def complete_multipart_upload(key, upload_id, etags, headers={}):
         query = urllib.parse.urlencode({'key': key})
@@ -227,22 +243,29 @@ def multipart_upload(local_path, companion_url, file_type, headers={}):
             part_size_mb = max(5, int(math.ceil(file_len_mb / 10000)))
             n_parts = int(math.ceil(file_len_mb / part_size_mb))
             while True:
+                semaphore.acquire()
                 file_data = f.read(part_size_mb * 1024 ** 2)
                 if not file_data:
                     break
 
                 part += 1
                 print(f'Uploading part {part:3}/{n_parts:3} of {Path(local_path).name} file...')
-                presigned_url = sign_part_upload(key, upload_id, part)
-                yield part, presigned_url, file_data
+                yield part, file_data
 
     session = requests.Session()
     key, upload_id = init_multipart_upload(Path(local_path).name, file_type, headers=headers)
 
-    with ThreadPoolExecutor(8) as ex:
+    # Python evaluates the input to ThreadPoolExecutor.map eagerly, which would allow iterate_file
+    # to read parts and sign uploads even if there was a significant queue to upload_part.
+    # This semaphore ensures that iterate_file can only read, sign and yield a part when upload_part
+    # is ready to receive it.
+    n_threads = 8
+    semaphore = BoundedSemaphore(n_threads)
+
+    with ThreadPoolExecutor(n_threads) as ex:
         etags = list(
             ex.map(
-                lambda args: (args[0], upload_part(*args[1:])),
+                lambda args: (args[0], upload_part(*args)),
                 iterate_file(key, upload_id, local_path),
             )
         )
@@ -1089,6 +1112,7 @@ class SMDataset(object):
         hotspot_clipping=False,
         neutral_loss='',
         chem_mod='',
+        image_metadata=[],
     ):
         """Retrieve ion images for a specific sf and adduct.
 
@@ -1104,18 +1128,9 @@ class SMDataset(object):
                                          This is required to get ion images that match the METASPACE website
         :param str   neutral_loss:
         :param str   chem_mod:
+        :param list  image_metadata:
         :return IsotopeImages:
         """
-        records = self._gqclient.getAnnotations(
-            {
-                'sumFormula': sf,
-                'adduct': adduct,
-                'databaseId': None,
-                'neutralLoss': neutral_loss,
-                'chemMod': chem_mod,
-            },
-            {'ids': self.id},
-        )
 
         import matplotlib.image as mpimg
 
@@ -1138,9 +1153,22 @@ class SMDataset(object):
             assert data.max() <= 1
             return data
 
-        if records:
-            image_metadata = records[0]['isotopeImages']
-        else:
+        if not image_metadata:
+            records = self._gqclient.getAnnotations(
+                {
+                    'sumFormula': sf,
+                    'adduct': adduct,
+                    'databaseId': None,
+                    'neutralLoss': neutral_loss,
+                    'chemMod': chem_mod,
+                },
+                {'ids': self.id},
+            )
+
+            if records:
+                image_metadata = records[0]['isotopeImages']
+
+        if not image_metadata:
             raise LookupError(f'Isotope image for "{sf}{chem_mod}{neutral_loss}{adduct}" not found')
         if only_first_isotope:
             image_metadata = image_metadata[:1]
@@ -1221,7 +1249,7 @@ class SMDataset(object):
         with ThreadPoolExecutor() as pool:
 
             def get_annotation_images(row):
-                sf, adduct, neutral_loss, chem_mod = row
+                sf, adduct, neutral_loss, chem_mod, isotope_images = row
                 return self.isotope_images(
                     sf,
                     adduct,
@@ -1230,18 +1258,23 @@ class SMDataset(object):
                     neutral_loss=neutral_loss,
                     chem_mod=chem_mod,
                     hotspot_clipping=hotspot_clipping,
+                    image_metadata=isotope_images,
                 )
 
             annotations = self.annotations(
                 fdr=fdr,
                 database=database,
-                return_vals=('sumFormula', 'adduct', 'neutralLoss', 'chemMod'),
+                return_vals=('sumFormula', 'adduct', 'neutralLoss', 'chemMod', 'isotopeImages'),
                 **annotation_filter,
             )
             return list(
-                pool.map(
-                    get_annotation_images,
-                    annotations,
+                tqdm(
+                    pool.map(
+                        get_annotation_images,
+                        annotations,
+                    ),
+                    total=len(annotations),
+                    bar_format='{l_bar}{bar:40}{r_bar}{bar:-10b}',
                 )
             )
 
@@ -1308,20 +1341,32 @@ class SMDataset(object):
                 ex.map(download_link, link['files'])
 
     def _get_diagnostic_image(self, image):
-        key = (image['url'], image['format'])
+        url = image['url']
+        fmt = image['format']
+        key = (url, fmt)
         if key not in self._diagnostic_images:
             try:
-                raw = BytesIO(self._session.get(image['url']).content)
+                raw = BytesIO(self._session.get(url).content)
             except Exception:
                 # Retry once in case of network error due to excessive parallelism
-                raw = BytesIO(self._session.get(image['url']).content)
+                raw = BytesIO(self._session.get(url).content)
+            try:
+                if fmt == 'PNG':
+                    import matplotlib.image as mpimg
 
-            if format == 'PNG':
-                import matplotlib.image as mpimg
-
-                image_content = mpimg.imread(raw)
-            else:
-                image_content = np.load(raw, allow_pickle=False)  # type: ignore
+                    image_content = mpimg.imread(raw)
+                elif fmt == 'NPY':
+                    image_content = np.load(raw, allow_pickle=False)  # type: ignore
+                elif fmt == 'JSON':
+                    image_content = json.load(raw)
+                elif fmt == 'PARQUET':
+                    image_content = pd.read_parquet(raw)
+                else:
+                    print(f'Warning: Unrecognized format {fmt}, returning unparsed content')
+                    image_content = raw
+            except Exception as ex:
+                print(f'Warning: Could not parse image {url}: {ex}')
+                image_content = raw
 
             self._diagnostic_images[key] = image_content
         return self._diagnostic_images[key]
@@ -1348,7 +1393,9 @@ class SMDataset(object):
         diagnostics = deepcopy(self._diagnostics)
 
         if include_images:
-            self._mixin_diagnostic_images(image for diag in diagnostics for image in diag['images'])
+            self._mixin_diagnostic_images(
+                [image for diag in diagnostics for image in diag.get('images') or []]
+            )
 
         return diagnostics
 
@@ -1449,11 +1496,50 @@ class SMInstance(object):
             from urllib3.exceptions import InsecureRequestWarning
 
             warnings.filterwarnings('ignore', category=InsecureRequestWarning)
-        self.reconnect()
+
+        try:
+            self.reconnect()
+        except (AssertionError, BadRequestException) as ex:
+            if ('Invalid API key' in ex.message or 'Login failed' in ex.message) and (
+                api_key is None and email is None
+            ):
+                print(
+                    f'Login failed. Call sm.save_login(overwrite=True) to update your '
+                    f'saved credentials.'
+                )
+            else:
+                print(f'Failed to connect to {self._config["host"]}: {ex.message}')
 
     def __repr__(self):
         return "SMInstance({})".format(self._config['graphql_url'])
 
+    def save_login(self, overwrite=False):
+        """Saves login credentials to the config file so that they will be automatically loaded
+        in future uses of SMInstance()"""
+        config_path = Path.home() / '.metaspace'
+        if config_path.exists() and not overwrite:
+            print(f'{config_path} already exists. Call sm.save_login(overwrite=True) to overwrite.')
+            return
+
+        api_key = getpass(
+            f'Please generate an API key at https://metaspace2020.eu/user/me and enter it here '
+            f'(or leave blank to cancel):'
+        )
+        api_key = api_key.strip()
+        if not api_key:
+            print('Cancelled')
+            return
+
+        try:
+            self.login(api_key=api_key)
+        except:
+            print(f'Login failed. Please check your API key and try again.')
+            return
+
+        config_path.open('w').write(f'api_key={api_key}')
+
+        print(f'Saved API key to {config_path}')
+
     def login(self, email=None, password=None, api_key=None):
         """
         DEPRECATED. Avoid calling this directly - pass credentials to SMInstance directly instead.
@@ -1765,7 +1851,7 @@ class SMInstance(object):
         e.g. to update a dataset's adducts:
 
         >>> sm.update_dataset(
-        >>>     dataset_id='2018-11-07_14h15m28s',
+        >>>     id='2018-11-07_14h15m28s',
         >>>     adducts=['[M]+', '+H', '+K', '+Na'],
         >>> )
 
@@ -1814,15 +1900,15 @@ class SMInstance(object):
         if chem_mods is not None:
             input_field['chemMods'] = chem_mods
         if is_public is not None:
-            input_field['is_public'] = is_public
+            input_field['isPublic'] = is_public
         if ppm is not None:
             input_field['ppm'] = ppm
         if num_isotopic_peaks is not None:
-            input_field['num_isotopic_peaks'] = num_isotopic_peaks
+            input_field['numPeaks'] = num_isotopic_peaks
         if decoy_sample_size is not None:
-            input_field['decoy_sample_size'] = decoy_sample_size
+            input_field['decoySampleSize'] = decoy_sample_size
         if analysis_version is not None:
-            input_field['analysis_version'] = analysis_version
+            input_field['analysisVersion'] = analysis_version
 
         try:
             self._gqclient.update_dataset(id, input_field, reprocess or False, force)
